/**********
 * 
 * diskmodel.c - 
 *     Model the behavior of an HP 97560 hard drive
 * 
 * Public interface:
 *   disk_init()
 *   faux_read()
 *   faux_write()
 * 
 * Nils Nieuwejaar 1995-96
 *
 *********/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include "diskmodel.h"

#define max(a,b) (((a) > (b)) ? (a) : (b))
#define diff(a,b) ((b.tv_sec - a.tv_sec) * 1000000 + (b.tv_usec - a.tv_usec))
#define cmp(a,b) ((a.tv_sec < b.tv_sec) ? -1 : \
		  ((a.tv_sec > b.tv_sec) ? 1 : (a.tv_usec - b.tv_usec)))
#define debug(a) 

static struct disk_region regions[NUM_REGIONS];
static struct disk_state ds;
static struct cache_state cs;
static struct timeval now;

static void
add_time(struct timeval *now, ulong_t us, struct timeval *then)
{
    then->tv_usec = now->tv_usec + us;
    then->tv_sec = now->tv_sec;
    
    while (then->tv_usec > 1000000) {
	then->tv_usec -= 1000000;
	then->tv_sec++;
    }
}

/*
 * Initialize the state of the disk.
 * Set up the logical->physical region mappings
 */
void
disk_init()
{
    int i;
    ulong_t t;

    regions[0].pstart = (TRACKS_PER_CYL + 4) * SEC_PER_TRACK;
    regions[0].pend = (646*TRACKS_PER_CYL + 4) * SEC_PER_TRACK - 1;
    
    regions[1].pstart = 654*TRACKS_PER_CYL * SEC_PER_TRACK;
    regions[1].pend = (1298*TRACKS_PER_CYL + 19) * SEC_PER_TRACK - 1;
    
    regions[2].pstart = 1308*TRACKS_PER_CYL * SEC_PER_TRACK;
    regions[2].pend = (1952*TRACKS_PER_CYL + 19) * SEC_PER_TRACK - 1;

    for (t=0, i=0; i<3; i++) {
	regions[i].length = regions[i].pend - regions[i].pstart + 1;
	regions[i].lstart = t;
	t += regions[i].length;
	regions[i].lend = t-1;
    }
    ds.ready_time.tv_usec = ds.ready_time.tv_sec = 0;
    ds.last_time.tv_usec = ds.last_time.tv_sec = 0;
    ds.add.cylinder = 0;
    ds.add.track = 0;
    ds.add.logical = ds.add.physical = 0;
    cs.valid = 0;
    gettimeofday(&(ds.last_time), NULL);
}

/*
 * Given a logical block number, translate it into a (cylinder,
 * head, sector) number
 */
static void
translate(ulong_t logical, struct disk_address *add)
{
    int r;
    int phys;

    r = 0;
    while (((int)logical - (int)(regions[r].lend)) > 0 && r<NUM_REGIONS)
	r++;

    debug(("%u %d %u\n", logical, r, (logical - regions[r].lstart)));
    phys = regions[r].pstart + (logical - regions[r].lstart);
    add->track = (ulong_t) (phys / (int) SEC_PER_TRACK);
    add->head = add->track % TRACKS_PER_CYL;
    add->cylinder = add->track / TRACKS_PER_CYL;
    add->logical = logical;
    add->physical = phys;
}

/*
 * How long will it take to move the head mechanism to the proper
 * cylinder and head?
 * 
 * 1) If the mechanism is already in place, we only have to change heads
 * 2) If we have to move a short distance, most of the time is spent
 *    accelerating.  Thus the time is proportional to the sqrt of the
 *    distance.
 * 3) Long seeks spend most of the time moving at a constant rate, after
 *    a (basically) constant time accellerating.
 */

static ulong_t
seek_time(struct disk_address *add)
{
    ulong_t cyl, tracks, us;

    cyl = (ulong_t) abs(add->cylinder - ds.add.cylinder);
    tracks = abs(ds.add.head - add->head);

    if (cyl == 0) {
	if (tracks)
	    us = 2500;
	else
	    us = 0;
    }
    else if (cyl < 383)
	us = 3240 + 400 * sqrt((double)cyl);
    else
	us = 8000 + 8 * cyl;
    
    debug(("Seektime.  cyl: %u  tracks: %u  us: %u\n", cyl, tracks, us));
    return(us);
}

/*
 * Once the head is positioned, how long does it take for the proper
 * sector to rotate under the head?
 */

static ulong_t
rot_delay(struct disk_address *add, struct timeval *start, ulong_t seektime)
{
    double revs, dest, rot;
    ulong_t us, past;
    int ss;

    past = diff(ds.last_time, (*start));

    /* Total revolutions since the last time we checked */
    revs = (double) (diff(ds.last_time, (*start)) + seektime) / USEC_PER_REV;

    /* Disregard whole revolutions - we're back where we started */
    revs = revs - (unsigned long) revs;

    ss = add->physical % SEC_PER_TRACK;
    dest = (double) (ss + add->cylinder * CYLINDERSKEW +
		     (add->track - add->cylinder) * TRACKSKEW) /
			 SEC_PER_TRACK;
    
    /* How far is it from where we are, to where we're going? */
    rot = 1 + dest - revs;
    rot = rot - (unsigned long) rot;
    
    us = seektime + (rot * USEC_PER_REV);
    debug(("rot_delay: us: %u\n", us-seektime));
    return(us-seektime);
}

/*
 * How long before the head is right over the data we want?
 */
static ulong_t
move_time(struct disk_address *add, struct timeval *start)
{
    ulong_t us, s, r;

    s = seek_time(add);
    r = rot_delay(add, start, s);

    debug(("   Seek: %u\tRot: %u\n", s, r));
    return(s+r);
}

/*
 * Move the head to the new address, and update the time
 */
static ulong_t
move_to(struct disk_address *add, ulong_t us)
{
    bcopy(add, &(ds.add), sizeof(struct disk_address));
    add_time(&now, us, &(ds.last_time));
}

/*
 * Once the head is in position, how long will it take to actually
 * transfer the data to disk?
 */
ulong_t
transfer_time(struct disk_address *add, int sectors)
{
    struct disk_address end;
    int cyl, track;

    translate(add->logical + sectors, &end);

    cyl = end.cylinder - add->cylinder;
    track = end.track - add->track - cyl;

    return(sectors * USEC_PER_SECTOR +
	   cyl * CYLINDER_SWITCH +
	   track * HEAD_SWITCH);
}

/*
 * Handle the disk's cache.  This gets pretty hairy.
 * Among the factors we need to handle: Is the data all the way
 * in the cache?  Are we currently loading the cache?  Are we loading
 * the data we want?
 * 
 * Basically, it boils down to 'when will the data be ready'?
 */
static long
in_cache(struct disk_address *add, int sectors)
{
    ulong_t first, last, done;
    ulong_t working, needed, ttime;
    int have;

    first = add->logical;
    last = add->logical + sectors;

    working = diff(cs.start, now);
    debug(("Start: %u Prefetch: %u Working: %u Want: %u\n",
	   cs.first.logical, cs.prefetch.logical, working,
	   add->logical));

    /* Is there any chance at all that this is in the cache?
     * If not, skip all this and start over */
    if (!cs.valid || (first < cs.first.logical) ||
	(first > cs.first.logical+CACHE_SLOTS)) {
	return(-1);
    }

    /* Is the whole thing already in the cache? */
    if (cmp(now, cs.done) >= 0 && last <= cs.first.logical+CACHE_SLOTS)
	return(0);

    /* How long has the disk been prefetching? */
    working = diff(cs.start, now);
    
    /* How long will it take to satisfy this request? */
    if (last < cs.prefetch.logical)
	needed = 0;
    else {
	/* If prefetching hasn't reached the start of this req, start over */
	if (first > cs.prefetch.logical &&
	    transfer_time(&cs.prefetch, first - cs.prefetch.logical) > working)
	{
	    debug(("Not here yet\n"));
	    return(-1);
	}
	needed = transfer_time(&cs.prefetch, last - cs.prefetch.logical);
    }

    /* If we aren't already prefetching, get it started */
    if (cmp(now, cs.done) >= 0) {
	have = cs.first.logical + CACHE_SLOTS - add->logical;
	if (have)
	    debug(("Had %d sectors\n", have));
	ttime = transfer_time(add, CACHE_SLOTS) - transfer_time(add, have);
	translate(add->logical + have, &cs.prefetch);
	add_time(&now, 0, &cs.start);
	add_time(&cs.start, ttime, &cs.done);
    }
    else {
	ttime = transfer_time(&cs.prefetch,
			      (first-cs.prefetch.logical)+CACHE_SLOTS);
	done = transfer_time(&cs.prefetch, first - cs.prefetch.logical);
	bcopy(add, &(cs.prefetch), sizeof(struct disk_address));
	add_time(&cs.start, done, &cs.start);
	add_time(&cs.start, ttime-done, &cs.done);
	debug(("total: %u  sofar: %u\n", ttime, done));
    }

    bcopy(add, &(cs.first), sizeof(struct disk_address));

    if (needed < working) {
	debug(("Ready now\n"));
	return(0);
    }
    debug(("   needed: %u working: %u left: %d\n",
	   needed, working, needed-working));
    return(needed - working);
}    

/*
 * Toss whatever is in the cache, and start loading our data.
 * Figure out when our data will be ready, and when the cache will be
 * fully loaded
 */
static void
reset_cache(struct disk_address *add, ulong_t wait)
{
    cs.valid = 1;
    bcopy(add, &(cs.first), sizeof(struct disk_address));
    bcopy(add, &(cs.prefetch), sizeof(struct disk_address));
    
    add_time(&now, wait, &cs.start);
    add_time(&cs.start, transfer_time(add, CACHE_SLOTS), &cs.done);
}

/*
 * Given an address and a number of sectors, how many microseconds
 * will it take to move that data from the disk across the bus to
 * the memory?
 */
ulong_t
read_time(struct disk_address *add, int sectors)
{
    ulong_t disk_time;   /* Time to transfer all data to disk */
    ulong_t find_time;   /* Time to position head for start of read */
    ulong_t wait_time;   /* Time until disk is ready to be to */
    ulong_t sector_bus_time; /* Time to transfer one sector across the bus */
    ulong_t total, tmp;
    long cache_wait;

    sector_bus_time = SECTOR_SIZE / BUSSPEED;

    debug(("Read:\tSec: %5u\t",  add->logical));

    cache_wait = in_cache(add, sectors);

    /* If the data is in the cache, send it right across the bus */
    if (!cache_wait)
	return((sectors * SECTOR_SIZE) / BUSSPEED);

    /* If the data is mostly in the cache, the bus-time will be the
     * dominant cost.  If only a small amount of the data is in the
     * cache, we'll spend most of our time waiting for the disk
     */
    if (cache_wait > 0)
	return(max(cache_wait + sector_bus_time,
		   (sectors * SECTOR_SIZE) / BUSSPEED));

    /* Nothing was in the cache.  We have to wait for any pending
     * writes to complete, then wait for the head to be positioned,
     * and then for our data to come off the disk.
     */
    disk_time = transfer_time(add, sectors);
    find_time = move_time(add, &now);

    if (cmp(now, ds.ready_time) >= 0)
	wait_time = 0;
    else
	wait_time = diff(now, ds.ready_time);
    
    total = wait_time + find_time + disk_time + sector_bus_time;

    move_to(add, wait_time + find_time);

    reset_cache(add, wait_time + find_time);
    return(total);
}

/*
 * Given an address and a number of sectors, how many microseconds
 * will it take to move that data across the bus to either cache or
 * disk (depending on synchronous flag).
 */
ulong_t
write_time(struct disk_address *add, int sectors, int sync)
{
    ulong_t disk_time;   /* Time to transfer all data to disk */
    ulong_t find_time;   /* Time to position head for start of write */
    ulong_t wait_time;   /* Time until disk is ready to be written to */
    ulong_t bus_time;    /* Time spent transferring data across bus */
    ulong_t pre_time;    /* Time spent transferring data to cache before
			  * writing to disk */
    ulong_t done_time;   /* When will all data be on disk? */
    ulong_t return_time; /* When do we return to caller? */
    int tmp, bytes;

    bytes = sectors * SECTOR_SIZE;
    bus_time = bytes / BUSSPEED;
    disk_time = transfer_time(add, sectors);
    tmp = cmp(now, ds.ready_time);

    /* If we get lucky, we can add this to the end of the data
     * already in the disk cache.
     */
    if (cs.valid == 2 && (cs.last+1) == add->logical && tmp<0) {
	cs.last == add->logical + sectors-1;
	add_time(&ds.ready_time, disk_time, &ds.ready_time);
	return(bus_time);
    }
    cs.last == add->logical + sectors-1;
    cs.valid = 2;

    /* Do we have to wait for the cache to be flushed? */
    if (tmp >= 0) {
	wait_time = 0;
	/* Cache is fine, start seeking */
	find_time = move_time(add, &now);
    }
    else {
	wait_time = diff(now, ds.ready_time);
	/* Can't seek until the cache is flushed */
 	find_time = move_time(add, &ds.ready_time);
    }

    debug(("   Now: (%u, %u)  Ready: (%u, %u)  Diff: %u\n",
	    now.tv_sec, now.tv_usec,
	    ds.ready_time.tv_sec, ds.ready_time.tv_usec, wait_time));

    /* We don't start writing immediately.  We wait until WRITEFENCE
     * bytes are ready for the disk.
     */
    if (bytes > (int) WRITEFENCE)
	pre_time = max(max(bus_time,find_time), (int) write_fence_time);
    else
	pre_time = max(bus_time,find_time);

    /* When will everything be on disk?  After the cache is flushed,
     * the data arrives across the bus, the disk head seeks, and
     * we actually write to disk.  Whew.
     */
    done_time = wait_time + pre_time + disk_time;

    debug(("   wait: %u  pre: %u  find: %u disk: %u done: %u\n",
	   wait_time, pre_time, find_time, disk_time, done_time));

    if (sync)
      return_time = done_time;
    else
      return_time = wait_time + bus_time;

    add_time(&now, done_time, &ds.ready_time);
    move_to(add, wait_time + find_time);

    return(return_time);
}

unsigned long
faux_read(ulong_t logical, int sectors)
{
    struct disk_address add;
    long usec;

    translate(logical, &add);
    gettimeofday(&now, NULL);

    /* encompasses controller overhead, grabbing the bus, etc */
    add_time(&now, OVERHEAD, &now);

    usec = OVERHEAD + read_time(&add, sectors);
    return(usec);
}

unsigned long
faux_write(ulong_t logical, int sectors, int immediate)
{
    struct disk_address add;
    long usec;

    translate(logical, &add);
    gettimeofday(&now, NULL);

    /* encompasses controller overhead, grabbing the bus, etc. */
    
    add_time(&now, OVERHEAD, &now);

    usec = write_time(&add, sectors, immediate);
    return(usec);
}
